@@ -1,5 +1,5 @@
 /*!
- * Convenient Formsets 0.9.3 (https://github.com/tiesjan/django-convenient-formsets/)
+ * Convenient Formsets 0.9.4 (https://github.com/tiesjan/django-convenient-formsets/)
  * Copyright 2021 Ties Jan Hefting
  * Licensed under BSD 3-Clause License
  */
@@ -50,6 +50,20 @@ const ConvenientFormset = function(options) {
             'defaultValue': undefined,
             'requiredIf': 'canDeleteForms',
         },
+
+        // Options for ordering forms
+        'canOrderForms': {
+            'defaultValue': false,
+            'requiredIf': 'never',
+        },
+        'moveFormDownButtonSelector': {
+            'defaultValue': undefined,
+            'requiredIf': 'canOrderForms',
+        },
+        'moveFormUpButtonSelector': {
+            'defaultValue': undefined,
+            'requiredIf': 'canOrderForms',
+        },
     };
 
     /* Formset specific options */
@@ -82,11 +96,42 @@ const ConvenientFormset = function(options) {
          * `for` attribute of labels and the `id`/`name` attributes of inputs.
          */
         const prefix = formsetOptions.formsetPrefix;
-        const idRegex = new RegExp('(' + prefix + '-(\\d+|__prefix__))');
+        const idRegex = new RegExp(prefix + '-(\\d+|__prefix__)');
 
-        const forms = formsetElements.formsContainer.querySelectorAll(
+        let forms = formsetElements.formsContainer.querySelectorAll(
             formsetOptions.formSelector
         );
+
+        // If forms can be ordered, order the retrieved forms by the `name`
+        // attribute of ORDER input elements. This ensures indexes are updated
+        // according to the server side's expectations.
+        if (formsetOptions.canOrderForms) {
+            const compareForms = function(leftForm, rightForm) {
+                const selector = 'input[name$="ORDER"]';
+                const leftFormOrderElement = leftForm.querySelector(selector);
+                const rightFormOrderElement = rightForm.querySelector(selector);
+
+                // Parse integral index from both form's ORDER input element
+                // name. New forms have an index value of '__prefix__', which
+                // cannot be parsed. They will moved to the back of the array.
+                const leftFormId = parseInt(
+                    leftFormOrderElement.name.match(idRegex)[1], 10
+                );
+                const rightFormId = parseInt(
+                    rightFormOrderElement.name.match(idRegex)[1], 10
+                );
+                if (isNaN(rightFormId) || leftFormId < rightFormId) {
+                    return -1;
+                }
+                else if (isNaN(leftFormId) || leftFormId > rightFormId) {
+                    return 1;
+                }
+            };
+
+            forms = Array.prototype.slice.call(forms).sort(compareForms);
+        }
+
+
         for (let i = 0; i < forms.length; i++) {
             const form = forms[i];
             const idReplacement = prefix + '-' + i;
@@ -95,10 +140,10 @@ const ConvenientFormset = function(options) {
             const labelElements = form.querySelectorAll('label');
             for (let j = 0; j < labelElements.length; j++) {
                 const labelElement = labelElements[j];
-                if (labelElement.hasAttribute('for')) {
-                    let attrValue = labelElement.getAttribute('for');
+                if (labelElement.htmlFor) {
+                    let attrValue = labelElement.htmlFor;
                     attrValue = attrValue.replace(idRegex, idReplacement);
-                    labelElement.setAttribute('for', attrValue);
+                    labelElement.htmlFor = attrValue;
                 }
             }
 
@@ -109,15 +154,15 @@ const ConvenientFormset = function(options) {
             );
             for (let j = 0; j < inputElements.length; j++) {
                 const inputElement = inputElements[j];
-                if (inputElement.hasAttribute('id')) {
-                    let attrValue = inputElement.getAttribute('id');
+                if (inputElement.id) {
+                    let attrValue = inputElement.id;
                     attrValue = attrValue.replace(idRegex, idReplacement);
-                    inputElement.setAttribute('id', attrValue);
+                    inputElement.id = attrValue;
                 }
-                if (inputElement.hasAttribute('name')) {
-                    let attrValue = inputElement.getAttribute('name');
+                if (inputElement.name) {
+                    let attrValue = inputElement.name;
                     attrValue = attrValue.replace(idRegex, idReplacement);
-                    inputElement.setAttribute('name', attrValue);
+                    inputElement.name = attrValue;
                 }
             }
         }
@@ -127,7 +172,7 @@ const ConvenientFormset = function(options) {
         /*
          * Updates the `TOTAL_FORMS` input in the management form to the number
          * of forms in the formset, regardless of whether it has been marked
-         * for deletion. Invoked after forms have been added or removed.
+         * for deletion. Invoked after forms have been added or deleted.
          */
         const allForms = formsetElements.formsContainer.querySelectorAll(
             formsetOptions.formSelector
@@ -170,6 +215,38 @@ const ConvenientFormset = function(options) {
             );
         }
 
+        // Add event listeners for clicks on the `moveFormDownButton` and
+        // `moveFormUpButton` if forms can be ordered, and set the initial
+        // value of the ORDER input element
+        if (formsetOptions.canOrderForms) {
+            const moveFormDownButton = newForm.querySelector(
+                    formsetOptions.moveFormDownButtonSelector);
+            moveFormDownButton.addEventListener(
+                'click', moveFormDownButtonClicked.bind(this, newForm)
+            );
+
+            const moveFormUpButton = newForm.querySelector(
+                    formsetOptions.moveFormUpButtonSelector);
+            moveFormUpButton.addEventListener(
+                'click', moveFormUpButtonClicked.bind(this, newForm)
+            );
+
+            const selector = 'input[name$="ORDER"]';
+            let newFormOrderValue;
+            if (visibleForms.length) {
+                const lastForm = visibleForms[visibleForms.length - 1];
+                const lastFormOrderValue = parseInt(
+                    lastForm.querySelector(selector).value, 10
+                );
+                newFormOrderValue = lastFormOrderValue + 1;
+            }
+            else {
+                newFormOrderValue = 1;
+            }
+            const newFormOrderElement = newForm.querySelector(selector);
+            newFormOrderElement.value = newFormOrderValue;
+        }
+
         // Append form to forms container
         formsetElements.formsContainer.appendChild(newForm);
 
@@ -186,16 +263,16 @@ const ConvenientFormset = function(options) {
     function deleteFormButtonClicked(form) {
         /*
          * Event handler for clicks on the `deleteFormButton`. Expects `form`
-         * as a bound argument. If a DELETE input field is present in the form,
-         * it's set to 'on' and the form is hidden. Otherwise the form is
-         * removed altogether from the DOM.
+         * as a bound argument. If a DELETE input element is present in the
+         * form, it's set to 'on' and the form is hidden. Otherwise the form is
+         * removed from the DOM altogether.
          *
          * In case the `hideAddFormButtonOnMaxForms` option is set to `true`,
          * it updates the visibility of the `addFormButton`.
          */
-        const deleteField = form.querySelector('input[name$="DELETE"]');
-        if (deleteField !== null) {
-            deleteField.value = 'on';
+        const deleteElement = form.querySelector('input[name$="DELETE"]');
+        if (deleteElement !== null) {
+            deleteElement.value = 'on';
             form.hidden = true;
         }
         else {
@@ -212,6 +289,62 @@ const ConvenientFormset = function(options) {
         updateManagementForm();
     }
 
+    function moveFormDownButtonClicked(form) {
+        /*
+         * Event handler for clicks on the `moveFormDownButton`. Expects `form`
+         * as a bound argument. Swaps both forms in the DOM and their ordering
+         * values.
+         */
+        let visibleForms = formsetElements.formsContainer.querySelectorAll(
+            formsetOptions.formSelector + ':not([hidden])'
+        );
+        visibleForms = Array.prototype.slice.call(visibleForms);
+        const formIndex = visibleForms.indexOf(form);
+
+        if (formIndex < visibleForms.length - 1) {
+            // Move form after next form
+            const nextForm = visibleForms[formIndex + 1];
+            formsetElements.formsContainer.insertBefore(nextForm, form);
+
+            // Select ORDER input elements for both forms and swap their values
+            const selector = 'input[name$="ORDER"]';
+            const formOrderElement = form.querySelector(selector);
+            const nextFormOrderElement = nextForm.querySelector(selector);
+            const formOrderValue = formOrderElement.value;
+            const nextFormOrderValue = nextFormOrderElement.value;
+            formOrderElement.value = nextFormOrderValue;
+            nextFormOrderElement.value = formOrderValue;
+        }
+    }
+
+    function moveFormUpButtonClicked(form) {
+        /*
+         * Event handler for clicks on the `moveFormUpButton`. Expects `form`
+         * as a bound argument. Swaps both forms in the DOM and their ordering
+         * values.
+         */
+        let visibleForms = formsetElements.formsContainer.querySelectorAll(
+            formsetOptions.formSelector + ':not([hidden])'
+        );
+        visibleForms = Array.prototype.slice.call(visibleForms);
+        const formIndex = visibleForms.indexOf(form);
+
+        if (formIndex > 0) {
+            // Move form before previous form
+            const previousForm = visibleForms[formIndex - 1];
+            formsetElements.formsContainer.insertBefore(form, previousForm);
+
+            // Select ORDER input elements for both forms and swap their values
+            const selector = 'input[name$="ORDER"]';
+            const formOrderElement = form.querySelector(selector);
+            const previousFormOrderElement = previousForm.querySelector(selector);
+            const formOrderValue = formOrderElement.value;
+            const previousFormOrderValue = previousFormOrderElement.value;
+            formOrderElement.value = previousFormOrderValue;
+            previousFormOrderElement.value = formOrderValue;
+        }
+    }
+
     /* Initialization functions */
     function initializeFormsetOptions(customOptions) {
         /*
@@ -237,7 +370,8 @@ const ConvenientFormset = function(options) {
             const optionRequired = (
                 (requiredIf === 'always') ||
                 (requiredIf === 'canAddForms' && formsetOptions.canAddForms) ||
-                (requiredIf === 'canDeleteForms' && formsetOptions.canDeleteForms)
+                (requiredIf === 'canDeleteForms' && formsetOptions.canDeleteForms) ||
+                (requiredIf === 'canOrderForms' && formsetOptions.canOrderForms)
             );
             if (optionRequired && typeof optionValue === 'undefined') {
                 missingOptions.push(optionKey);
@@ -263,7 +397,8 @@ const ConvenientFormset = function(options) {
     function checkEmptyFormElements() {
         /*
          * Asserts that the empty form has an `deleteFormButton` if forms can
-         * be deleted.
+         * be deleted, and a `moveFormDownButton` and a `moveFormUpButton` if
+         * forms can be ordered.
          */
         const missingElements = [];
         let element;
@@ -277,6 +412,26 @@ const ConvenientFormset = function(options) {
             }
         }
 
+        if (formsetOptions.canOrderForms) {
+            selector = formsetOptions.moveFormDownButtonSelector;
+            element = formsetElements.emptyForm.querySelector(selector);
+            if (element === null) {
+                missingElements.push(selector);
+            }
+
+            selector = formsetOptions.moveFormUpButtonSelector;
+            element = formsetElements.emptyForm.querySelector(selector);
+            if (element === null) {
+                missingElements.push(selector);
+            }
+
+            selector = 'input[name$="ORDER"]';
+            element = formsetElements.emptyForm.querySelector(selector);
+            if (element === null) {
+                missingElements.push(selector);
+            }
+        }
+
         // Throw error if DOM elements are missing
         if (missingElements.length) {
             const formattedMissingElements = missingElements.map(
@@ -293,7 +448,8 @@ const ConvenientFormset = function(options) {
     function checkVisibleFormsElements() {
         /*
          * Asserts that all visible forms have a `deleteFormButton` if forms
-         * can be deleted.
+         * can be deleted, and a `moveFormDownButton` and a `moveFormUpButton`
+         * if forms can be ordered.
          */
         const missingElements = [];
         let element;
@@ -303,14 +459,35 @@ const ConvenientFormset = function(options) {
             formsetOptions.formSelector + ':not([hidden])'
         );
         for (let i = 0; i < forms.length; i++) {
+            const form = forms[i];
+
             if (formsetOptions.canDeleteForms) {
-                const form = forms[i];
                 selector = formsetOptions.deleteFormButtonSelector;
                 element = form.querySelector(selector);
                 if (element === null && missingElements.indexOf(selector) === -1) {
                     missingElements.push(selector);
                 }
             }
+
+            if (formsetOptions.canOrderForms) {
+                selector = formsetOptions.moveFormDownButtonSelector;
+                element = form.querySelector(selector);
+                if (element === null && missingElements.indexOf(selector) === -1) {
+                    missingElements.push(selector);
+                }
+
+                selector = formsetOptions.moveFormUpButtonSelector;
+                element = form.querySelector(selector);
+                if (element === null && missingElements.indexOf(selector) === -1) {
+                    missingElements.push(selector);
+                }
+
+                selector = 'input[name$="ORDER"]';
+                element = form.querySelector(selector);
+                if (element === null && missingElements.indexOf(selector) === -1) {
+                    missingElements.push(selector);
+                }
+            }
         }
 
         // Throw error if DOM elements are missing
@@ -425,8 +602,8 @@ const ConvenientFormset = function(options) {
         );
         for (let i = 0; i < forms.length; i++) {
             const form = forms[i];
-            const deleteField = form.querySelector('input[name$="DELETE"]');
-            if (deleteField !== null && deleteField.value === 'on') {
+            const deleteElement = form.querySelector('input[name$="DELETE"]');
+            if (deleteElement !== null && deleteElement.value === 'on') {
                 form.hidden = true;
             }
         }
@@ -435,7 +612,8 @@ const ConvenientFormset = function(options) {
     function initializeEventListeners() {
         /*
          * Initializes click event listeners for the `addFormButton` and for
-         * the `deleteFormButton` of visible forms.
+         * the `deleteFormButton`, `moveFormDownButton` and `moveFormUpButton`
+         * of visible forms.
          */
         if (formsetOptions.canAddForms) {
             formsetElements.addFormButton.addEventListener(
@@ -446,15 +624,30 @@ const ConvenientFormset = function(options) {
         const forms = formsetElements.formsContainer.querySelectorAll(
             formsetOptions.formSelector + ':not([hidden])'
         );
-        if (formsetOptions.canDeleteForms) {
-            for (let i = 0; i < forms.length; i++) {
-                const form = forms[i];
+        for (let i = 0; i < forms.length; i++) {
+            const form = forms[i];
+
+            if (formsetOptions.canDeleteForms) {
                 const deleteFormButton = form.querySelector(
                         formsetOptions.deleteFormButtonSelector);
                 deleteFormButton.addEventListener(
                     'click', deleteFormButtonClicked.bind(this, form)
                 );
             }
+
+            if (formsetOptions.canOrderForms) {
+                const moveFormDownButton = form.querySelector(
+                        formsetOptions.moveFormDownButtonSelector);
+                moveFormDownButton.addEventListener(
+                    'click', moveFormDownButtonClicked.bind(this, form)
+                );
+
+                const moveFormUpButton = form.querySelector(
+                        formsetOptions.moveFormUpButtonSelector);
+                moveFormUpButton.addEventListener(
+                    'click', moveFormUpButtonClicked.bind(this, form)
+                );
+            }
         }
     }
 
